@@ -17,7 +17,7 @@
 use serde::{Deserialize, Serialize};
 
 /// Returned value for the `getnodestats` rpc call
-#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Debug, Serialize, Deserialize)]
 pub struct NodeStats {
     /// Stats related to messages received by the node.
     pub inbound: NodeInboundStats,
@@ -31,6 +31,8 @@ pub struct NodeStats {
     pub queues: NodeQueueStats,
     /// Miscellaneous stats related to the node.
     pub misc: NodeMiscStats,
+    /// The node's internal RTT from message received to response sent.
+    pub internal_rtt: NodeInternalRttStats,
 }
 
 #[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
@@ -75,7 +77,7 @@ pub struct NodeOutboundStats {
     pub all_failures: u64,
 }
 
-#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Debug, Serialize, Deserialize)]
 pub struct NodeConnectionStats {
     /// The number of all connections the node has accepted.
     pub all_accepted: u64,
@@ -83,6 +85,8 @@ pub struct NodeConnectionStats {
     pub all_initiated: u64,
     /// The number of rejected inbound connection requests.
     pub all_rejected: u64,
+    /// The average connection duration (in seconds).
+    pub average_duration: f64,
     /// Number of currently connecting peers.
     pub connecting_peers: u32,
     /// Number of currently connected peers.
@@ -109,18 +113,28 @@ pub struct NodeHandshakeStats {
 
 #[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
 pub struct NodeQueueStats {
+    /// The number of queued consensus requests.
+    pub consensus: u64,
     /// The number of messages queued in the common inbound channel.
     pub inbound: u64,
     /// The number of messages queued in the individual outbound channels.
     pub outbound: u64,
+    /// The number of queued peer events.
+    pub peer_events: u64,
+    /// The number of queued storage requests.
+    pub storage: u64,
+    /// The number of queued sync items.
+    pub sync_items: u64,
 }
 
-#[derive(Clone, Debug, Eq, PartialEq, Serialize, Deserialize)]
+#[derive(Clone, Debug, Serialize, Deserialize)]
 pub struct NodeMiscStats {
     /// The current block height of the node.
     pub block_height: u64,
     /// The number of blocks the node has mined.
     pub blocks_mined: u64,
+    /// The average block processing time (in seconds).
+    pub block_processing_time: f64,
     /// The number of duplicate blocks received.
     pub duplicate_blocks: u64,
     /// The number of duplicate sync blocks received.
@@ -130,3 +144,15 @@ pub struct NodeMiscStats {
     /// The number of RPC requests received.
     pub rpc_requests: u64,
 }
+
+#[derive(Clone, Debug, Serialize, Deserialize)]
+pub struct NodeInternalRttStats {
+    /// The average internal RTT for `GetPeer` queries (in seconds).
+    pub getpeers: f64,
+    /// The average internal RTT for `GetSync` queries (in seconds).
+    pub getsync: f64,
+    /// The average internal RTT for `GetBlocks` queries (in seconds).
+    pub getblocks: f64,
+    /// The average internal RTT for `GetMemoryPool` queries (in seconds).
+    pub getmemorypool: f64,
+}
