@@ -17,12 +17,13 @@
 use metrics::{GaugeValue, Key, Recorder, Unit};
 
 use crate::{
-    metric_types::{Counter, DiscreteGauge},
+    metric_types::{CircularHistogram, Counter, DiscreteGauge},
     names::*,
     snapshots::{
         NodeConnectionStats,
         NodeHandshakeStats,
         NodeInboundStats,
+        NodeInternalRttStats,
         NodeMiscStats,
         NodeOutboundStats,
         NodeQueueStats,
@@ -45,6 +46,8 @@ pub struct Stats {
     queues: QueueStats,
     /// Miscellaneous stats related to the node.
     misc: MiscStats,
+    /// The node's internal RTT from message received to response sent (in seconds).
+    internal_rtt: InternalRtt,
 }
 
 impl Stats {
@@ -56,6 +59,7 @@ impl Stats {
             handshakes: HandshakeStats::new(),
             queues: QueueStats::new(),
             misc: MiscStats::new(),
+            internal_rtt: InternalRtt::new(),
         }
     }
 
@@ -67,6 +71,7 @@ impl Stats {
             handshakes: self.handshakes.snapshot(),
             queues: self.queues.snapshot(),
             misc: self.misc.snapshot(),
+            internal_rtt: self.internal_rtt.snapshot(),
         }
     }
 }
@@ -151,6 +156,8 @@ pub struct OutboundStats {
     all_successes: Counter,
     /// The number of messages that failed to be sent to peers.
     all_failures: Counter,
+    /// The number of messages that were going to be sent to a peer, but was blocked at the last minute by a cache.
+    all_cache_hits: Counter,
 }
 
 impl OutboundStats {
@@ -158,6 +165,7 @@ impl OutboundStats {
         Self {
             all_successes: Counter::new(),
             all_failures: Counter::new(),
+            all_cache_hits: Counter::new(),
         }
     }
 
@@ -182,6 +190,8 @@ pub struct ConnectionStats {
     connected_peers: DiscreteGauge,
     /// Number of known disconnected peers.
     disconnected_peers: DiscreteGauge,
+    /// Tracks connection durations (once closed).
+    duration: CircularHistogram,
 }
 
 impl ConnectionStats {
@@ -193,6 +203,7 @@ impl ConnectionStats {
             connecting_peers: DiscreteGauge::new(),
             connected_peers: DiscreteGauge::new(),
             disconnected_peers: DiscreteGauge::new(),
+            duration: CircularHistogram::new(),
         }
     }
 
@@ -201,6 +212,7 @@ impl ConnectionStats {
             all_accepted: self.all_accepted.read(),
             all_initiated: self.all_initiated.read(),
             all_rejected: self.all_rejected.read(),
+            average_duration: self.duration.average(),
             connecting_peers: self.connecting_peers.read() as u32,
             connected_peers: self.connected_peers.read() as u32,
             disconnected_peers: self.disconnected_peers.read() as u32,
@@ -248,24 +260,40 @@ impl HandshakeStats {
 }
 
 pub struct QueueStats {
-    /// The number of messages queued in the common inbound channel.
+    /// The number of queued consensus items.
+    consensus: DiscreteGauge,
+    /// The number of messages queued in the individual inbound channels.
     inbound: DiscreteGauge,
     /// The number of messages queued in the individual outbound channels.
     outbound: DiscreteGauge,
+    /// The number of queued peer events.
+    peer_events: DiscreteGauge,
+    /// The number of queued storage requests.
+    storage: DiscreteGauge,
+    /// The number of queued sync items.
+    sync_items: DiscreteGauge,
 }
 
 impl QueueStats {
     const fn new() -> Self {
         Self {
+            consensus: DiscreteGauge::new(),
             inbound: DiscreteGauge::new(),
             outbound: DiscreteGauge::new(),
+            peer_events: DiscreteGauge::new(),
+            storage: DiscreteGauge::new(),
+            sync_items: DiscreteGauge::new(),
         }
     }
 
     pub fn snapshot(&self) -> NodeQueueStats {
         NodeQueueStats {
+            consensus: self.consensus.read(),
             inbound: self.inbound.read(),
             outbound: self.outbound.read(),
+            peer_events: self.peer_events.read(),
+            storage: self.storage.read(),
+            sync_items: self.sync_items.read(),
         }
     }
 }
@@ -274,6 +302,8 @@ pub struct MiscStats {
     block_height: DiscreteGauge,
     /// The number of mined blocks.
     blocks_mined: Counter,
+    /// The processing time for a block.
+    block_processing_time: CircularHistogram,
     /// The number of duplicate blocks received.
     duplicate_blocks: Counter,
     /// The number of duplicate sync blocks received.
@@ -289,6 +319,7 @@ impl MiscStats {
         Self {
             block_height: DiscreteGauge::new(),
             blocks_mined: Counter::new(),
+            block_processing_time: CircularHistogram::new(),
             duplicate_blocks: Counter::new(),
             duplicate_sync_blocks: Counter::new(),
             orphan_blocks: Counter::new(),
@@ -300,6 +331,7 @@ impl MiscStats {
         NodeMiscStats {
             block_height: self.block_height.read(),
             blocks_mined: self.blocks_mined.read(),
+            block_processing_time: self.block_processing_time.average(),
             duplicate_blocks: self.duplicate_blocks.read(),
             duplicate_sync_blocks: self.duplicate_sync_blocks.read(),
             orphan_blocks: self.orphan_blocks.read(),
@@ -308,6 +340,35 @@ impl MiscStats {
     }
 }
 
+/// Each histogram holds the last `QUEUE_CAPACITY` (see `metric_types` mod) measurements for internal RTT for the indicated message
+/// type. The snapshot produced for the RPC stats is the average RTT for each set.
+pub struct InternalRtt {
+    getpeers: CircularHistogram,
+    getsync: CircularHistogram,
+    getblocks: CircularHistogram,
+    getmemorypool: CircularHistogram,
+}
+
+impl InternalRtt {
+    const fn new() -> Self {
+        Self {
+            getpeers: CircularHistogram::new(),
+            getsync: CircularHistogram::new(),
+            getblocks: CircularHistogram::new(),
+            getmemorypool: CircularHistogram::new(),
+        }
+    }
+
+    pub fn snapshot(&self) -> NodeInternalRttStats {
+        NodeInternalRttStats {
+            getpeers: self.getpeers.average(),
+            getsync: self.getsync.average(),
+            getblocks: self.getblocks.average(),
+            getmemorypool: self.getmemorypool.average(),
+        }
+    }
+}
+
 impl Recorder for Stats {
     // The following are unused in Stats
     fn register_counter(&self, _key: &Key, _unit: Option<Unit>, _desc: Option<&'static str>) {}
@@ -316,7 +377,19 @@ impl Recorder for Stats {
 
     fn register_histogram(&self, _key: &Key, _unit: Option<Unit>, _desc: Option<&'static str>) {}
 
-    fn record_histogram(&self, _key: &Key, _value: f64) {}
+    fn record_histogram(&self, key: &Key, value: f64) {
+        let metric = match key.name() {
+            connections::DURATION => &self.connections.duration,
+            misc::BLOCK_PROCESSING_TIME => &self.misc.block_processing_time,
+            internal_rtt::GETPEERS => &self.internal_rtt.getpeers,
+            internal_rtt::GETSYNC => &self.internal_rtt.getsync,
+            internal_rtt::GETBLOCKS => &self.internal_rtt.getblocks,
+            internal_rtt::GETMEMORYPOOL => &self.internal_rtt.getmemorypool,
+            _ => return,
+        };
+
+        metric.push(value);
+    }
 
     fn increment_counter(&self, key: &Key, value: u64) {
         let metric = match key.name() {
@@ -339,6 +412,7 @@ impl Recorder for Stats {
             // outbound
             outbound::ALL_SUCCESSES => &self.outbound.all_successes,
             outbound::ALL_FAILURES => &self.outbound.all_failures,
+            outbound::ALL_CACHE_HITS => &self.outbound.all_cache_hits,
             // connections
             connections::ALL_ACCEPTED => &self.connections.all_accepted,
             connections::ALL_INITIATED => &self.connections.all_initiated,
@@ -368,6 +442,8 @@ impl Recorder for Stats {
             // queues
             queues::INBOUND => &self.queues.inbound,
             queues::OUTBOUND => &self.queues.outbound,
+            queues::PEER_EVENTS => &self.queues.peer_events,
+            queues::STORAGE => &self.queues.storage,
             // misc
             misc::BLOCK_HEIGHT => &self.misc.block_height,
             // connections
